// HFPolyMesh.h: interface for the HFPolyMesh class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(HFPOLYMESH1_H)
#define HFPolyMesh_H

// #pragma warning( disable : 4786 )

#include <vector>
#include <iostream>
#include <map>
#include <string>

#include <cmath>  


using namespace std;

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#define F1_0 0			// 0000000000000000
#define F1_1 1			// 0000000000000001
#define F1_2 2			// 0000000000000010
#define F1_4 3			// 0000000000000011

#define F2_0 0			// 0000000000000000
#define F2_1 4			// 0000000000000100
#define F2_2 8			// 0000000000001000
#define F2_4 12			// 0000000000001100

#define F3_0 0			// 0000000000000000
#define F3_1 16			// 0000000000010000
#define F3_2 32			// 0000000000100000
#define F3_4 48			// 0000000000110000

#define F4_0 0			// 0000000000000000
#define F4_1 64			// 0000000001000000
#define F4_2 128		// 0000000010000000
#define F4_4 192		// 0000000011000000

#define F5_0 0			// 0000000000000000
#define F5_1 256		// 0000000100000000
#define F5_2 512		// 0000001000000000
#define F5_4 768		// 0000001100000000

#define F6_0 0			// 0000000000000000
#define F6_1 1024		// 0000010000000000
#define F6_2 2048		// 0000100000000000
#define F6_4 3072		// 0000110000000000

#define NONE 0			// 0000000000000000
#define One	1			// 0000000000000001
#define TWO 2			// 0000000000000010
#define FOUR 3			// 0000000000000011

const int Face2Vert[6][4]={ {0,1,2,3}, {0,9,4,8}, {3,11,7,8}, {4,5,6,7}, {2,10,6,11}, {1,10,5,9} };
//{{0,1,2,3}, {3,11,7,8}, {4,5,6,7}, {1,10,5,9}, {0,9,4,8}, {2,10,6,11}};

const vector<double> EMPTYVD(0);
const vector<unsigned int> EMPTYVI(0);

#define EPS (double)0.00001
#define NORMDELTMULT (double)0.001
//#define pol_eps	0.001

struct HFPError{ string error;
	HFPError(string e){ error=e; }
	};

class HFPolyMesh{
	public:
//**********   PUBLIC METHODS   **********
		//HFPolyMesh can throw HFPError
		HFPolyMesh(double (* EXT_FUNTION) (vector<double>,double *));
		virtual ~HFPolyMesh();

		void Calc();
		
		int VertexNum();
		int TriangleNum();
		
		void Timer(bool t);
		bool Timer();

		void Normals(bool s);
		bool Normals();

		void SearchPercent(double p);
		double SearchPercent();
		void Search(bool s);
		bool Search();

		void Reduce(bool r);
		bool Reduce();
		
		void IsoValue(double isov);
		double IsoValue();

		void MinMax(vector<double> mm);
		vector<double> MinMax();

		void Grid(vector<int> g);
		vector<int> Grid();

		void DimMap(vector<int> dm);
		vector<int> DimMap();

		void Constants(vector<double> con);
		vector<double> Constants();

		void Parameters(vector<double> par);
		vector<double> Parameters();
		
		void Refine(int);
		int Refine();


		class TriData{
			public:
				TriData();
				~TriData();
				//Initilization 
			
				/*********ADDED FOR ATTRIBUTES*/
				/*----Added for refinement---*/
				void ReplaceVertex(int, vector<double>,double,double *);
				void setAttributes(int, double *);

				//void guessSize(int tNum, int );
				void guessSize(int tNum, int VNum,int );
				//Data Creation
				int pushBackVertex(double* x, double fvalue,double *S);
				int pushBackTriangle(int v[3]);
				//Data Modification
				//void setVertex(int v, int x, double coord,double *);
				//void setVertex(int v, double coords[3],double *);
				//void setVertex(int v, double coords[3], double fvalue,double *);
				//void setVertex(int v, double coords[3], double fvalue, double normal[3],double *);
				void initNormal();
				void setNormal(int v, double n1, double n2, double n3);
				
				/** Modified by schmitt :03 May 2001
				 * Call inside the hfpolymesh class. 
				 * No need to use a double *, and copy it into a vector data type
				 */
				void setNormal(int , vector<double>);
				
				void setFunctVal(int v, double fvalue);
				void deleteEdge(int e);
				void refresh();
				void ccwTriangles(); //Requires Normals
				//Data Retrival
				int getNumV();
				int getNumT();
				vector<double> getVertex(int v);
				double getVertex(int v, int x);
				vector<double> getNormal(int v);
				double getNormal(int v, int x);
				double getFunctVal(int v);
				vector<int> getTriangle(int t);
				int getTriangle(int t, int vnum);//vnum is 0, 1, or 2 only

				/*****ADDED FOR ATTRIBUTES***********/
				vector<double> getAttributes(int v);
				double getAttributes(int v, int x);
				double getVal(int);

				
				//For Later
				//int breakEdge(int e);
				//int breakTriangle3(int t);
				//int breakTriangle4(int t);
				//AND OTHERS

			private:
				//Vertex Data
				vector<vector<double> > itsVCoordinates;//Size 3
				vector<vector<int> > itsVEdges;//Size Variable
				vector<bool> itsVBad;
				vector<double> itsVFVal;
				vector<vector<double> > itsVNormal;//Size 3
				//Edge Data
				vector<vector<int> > itsEVertices;//Size 2
				vector<vector<int> > itsETriangles;//Size 2
				vector<bool> itsEBad;
				vector<double> itsELength;
				//Triangle Data
				vector<vector<int> > itsTEdges;//Size 3
				vector<bool> itsTBad;
				vector<bool> itsTRev;

				/*****ADDED FOR ATTRIBUTES */
				//Attributes Data
				vector<vector<double> > itsSAttributes;//Size itsSSize
				int itsSSize;
			
				int itsVBadCount;
				int itsEBadCount;
				int itsTBadCount;

				vector<int> itsOpenEdges;

				void fillDoubleVector(double* d, int size, vector<double>& ret);
				void fillIntVector(int* d, int size, vector<int>& ret);
				double edgeLength(int v1, int v2);
				void refreshTriangles();
				void refreshEdges();
				void refreshVerticies();
			};

		TriData itsData;


		
	private:
//**********   PRIVATE METHODS   **********
		void TimerStart(string tString);
		void TimerStop();
		void FillMatrix();
		void FillVertices();
		void CreateTriangles();
		void CreateNormals();

	/* ADDED FOR ATTRIBUTES REFINEMENT*/		
		void Refinement();
		int FindPositiveValue1(vector<double> *, vector<double>, vector<double>,vector<double> );
		int FindPositiveValue2(vector<double> *vertex,vector<double> step,vector< vector<double> > cube,vector<double>);
		void CalcNormal(int i,vector<double> *norm);
		
		/***************************************/
		
		double VertEst(const double& left, const double& right, const double& delta);
		double EdgeSearch(int leftI, int leftJ, int leftK, const int& select, const double lp[3], double* centerPos);

		double CalcVal(const double pos[3]);
		
		void VertCalc(int i, int j, int k, int select, const double lNodePos[3]);

		void StoreCell(int i, int j, int k, int select);

		void GraphBuildCase2(const int* Edge, int* I[12]);
		void GraphBuildCase4(double V,double A1, double A2, const int* Edge);

		void TraceGraph(int* I[12]);

		void FillIndex(int i, int j, int k, int* I[12]);

		void AddFace1ToCell(int li, int lj, int lk);
		void AddFace2ToCell(int li, int lj, int lk);
		void AddFace3ToCell(int li, int lj, int lk);
		void AddFace4ToCell(int li, int lj, int lk);
		void AddFace5ToCell(int li, int lj, int lk);
		void AddFace6ToCell(int li, int lj, int lk);

//**********   PRIVATE MEMBERS   **********
		bool itsSearch;			
		double itsSearchPer;

		bool itsReduce;

		bool itsNormals;
		bool itsNormalDisplay;

		bool itsTimer;
		double itsS0;					//*** Iso Value for Surface ***
		short itsXMap, itsYMap, itsZMap;//*** Its X,Y,Z Mapping
		double itsBBMax[3], itsBBMin[3];//*** Bounding Box Definition ***
		int itsGridSize[3];				//*** Grid Size Definition ***
		vector<double> itsSendingX;
		double itsEPS;					//*** Epsilon Value ***
		double itsDelta[3];				//*** Delta Values ***

		vector<double> itsNormDelt;

		double itsMinEdgeLen;
		int itsFirstTriangleInSlice;
		vector<vector<vector<int> > > DupVert;
		
		vector< vector < vector <double> > > itsValGrid;		//*** 3D Grid Holding Function Value Points ***
		vector< vector < vector <bool> > > itsBoolGrid;			//*** 3D Grid Holding Boolean for Inside/Outside Description of Grid Points ***
		vector< vector< vector < vector <int> > > > itsVertGrid;//*** 3D Grids Holding Integer Vertex indeces ***
		vector< vector < vector <int> > > itsCellGrid;			//*** 3D Grid Holding bitmap Style Info about Cell Faces ***
		vector< vector <int> > itsConnect;						//*** Conectivity Graph ***

		/*****ADDED FOR ATTRIBUTES */
		double * itsSAttributes;
		int itsSSize;
		int itsRefinement;


		/**************************/
			double (*calc) (vector<double>,double *);

};

#endif // !defined(HFPolyMesh_H)
